/*
 * 2012-06-24 12:38 AM
 * San Jose, California USA
 *
 * burplex photographer productivity tool - make image selects
 *
 * Copyright (c) 2012, Waitman Gobble <waitman@waitman.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Documentation +.v.+
 *
 * This program is intended for photographers who copy images from their memory cards
 * to their home directory. The photographer may browse the directory structure and
 * quickly review the images, marking selects or other purposes such as preview images
 * with a model, for example. Selects are stored in a file which may be used for batch
 * process scripting. Burplex will display images stored in RAW, JPEG, or PPM format.
 *
 * Note: check 'file' (/usr/bin/file) command output for your raw identifier. Output
 * should match the string 'raw image data'. If not, then you may wish to update the
 * corresponding line in the code below.
 *
 * The tree may be navigated with the mouse, or the up and down arrow keys.
 * To select a photo, press the Insert key. If you prefer a different key,
 * see the instructions below.
 *
 * When a select is made, the file '.selects', located in the same directory as the
 * image, is updated and saved. Pressing the Insert key a second time will remove the
 * select status, and update the '.selects' file (removing the image).
 *
 * The '.selects' file is useful for batch process scripting.
 *
 * Example 1: (bash, copy to 'sav' directory)
 * # while read -r line; do cp $line sav/; done < ".selects"
 *
 * Example 2: (bash, ImageMagick convert to resize)
 * # while read -r line; do convert -size 1200x1200 -scale 1200x1200 -density 120 -quality 90 $line sav/$line; done < ".selects"
 *
 * Note: star.png is stored in images/star.png in the source distribution.
 * It is specified in the burplex.rc RCC Resource File.
 */

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QLabel>
#include <QString>
#include <QFileDialog>
#include <QGraphicsView>
#include <QGraphicsScene>
#include <QGridLayout>
#include <qmath.h>


MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    /* load resource file: contains star.png
     * http://openclipart.org/detail/14693/emblem-star-by-rocket000
     */
    QResource::registerResource("burplex.rc");

    ui->setupUi(this);

    /* horizontal split - left is tree right is graphics view */
    QSplitter *splitter = new QSplitter(ui->centralWidget);

    gv = new QGraphicsView(this);
    gs = new QGraphicsScene(this);

    tree = new QTreeView;

    splitter->setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
    splitter->addWidget(tree);
    splitter->addWidget(gv);
    gv->show();

    /* set treeview model to file system */
    model = new QFileSystemModel;
    /* set root to home directory */
    model->setRootPath(QDir::homePath());
    /* sort by name */
    model->sort(0,Qt::AscendingOrder);

    /* sort keys:
     * 0 - Name
     * 1 - Size
     * 2 - Type
     * 3 - Date Modified
     */
    tree->setModel(model);
    tree->setAnimated(false);
    tree->setSortingEnabled(true);
    /* sort by Name
     * see sort keys above
     */
    tree->sortByColumn(0, Qt::AscendingOrder);
    tree->setColumnWidth(0, 250);
    tree->setSelectionMode(QAbstractItemView::SingleSelection);

    /* set root to user home directory
     * if you comment the following line, root will be /
     */
    tree->setRootIndex(model->index(QDir::homePath()));

    /* connect signal to load image when tree selection changes
     * tree selection can be changed with mouse or up/down arrow keys
     */
    connect( tree->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)),
             this, SLOT( changeCurrent(QItemSelection,QItemSelection)), Qt::DirectConnection ) ;

    setCentralWidget(splitter);
}

/* this is called when the selection changes in tree
 * load selects for path directory
 */
void MainWindow::changeCurrent(const QItemSelection &, const QItemSelection & )
{

    path = "";
    isRaw = false;

    /* get the path of the currently selected item
     * save in global var 'path'
     */
    QModelIndexList list =  tree->selectionModel()->selectedIndexes();

    foreach (QModelIndex index, list)
    {
        if (path=="")
        {
            path = model->filePath(index);
        }
    }

    if (path!="") {
        /* use the file command to check if supported image
         * ppm, jpeg, RAW
         * set to path to file command on your system
         * FreeBSD should be /usr/bin/find
         */
        QStringList arguments;
        QString foo = "/usr/bin/file "+path+"";
        arguments << "-c" << foo;
        QProcess* process = new QProcess(this);

        /* run process through /bin/sh
         * change to match your system
         * FreeBSD should be /bin/sh
         */
        process->start("/bin/sh", arguments);
        process->waitForFinished();

        /* load results of file command into string tmp
         * if JPEG or Netpbm (PPM) is seen then handle appropriately
         */
        QString tmp = process->readAll();
        if ((tmp.lastIndexOf("JPEG")>0)||(tmp.lastIndexOf("Netpbm")>0)) {
            /* update window title with path to image file */
            setWindowTitle("burplex:"+path);
            /* if this path is different than previous selected path,
             * load selects for this path
             */
            QString chkpath = getSelectPath(path);
            if (chkpath!=selectpath) {
                selectpath = chkpath;
                loadSelects();
            }
            /* load and display image. if this is a select then show star */
            loadImage(path,selects.contains(path));
         } else {
            /* if results of file command return 'raw image data'
             * then we have a RAW image from the camera
             * adjust to match your system
             * ex: file /path/to/raw/image.CR2
             * RAW images set global var isRaw = true
             */
            if (tmp.lastIndexOf("raw image data")) {
                isRaw=true;
                setWindowTitle("burplex:"+path);
                QString chkpath = getSelectPath(path);
                if (chkpath!=selectpath) {
                    selectpath = chkpath;
                    loadSelects();
                }
                /* extract and display thumbnail from raw image, jpeg format */
                QString thumbpath = getThumbnailPath(path);
                loadImage(thumbpath,selects.contains(path));
            }
        }
    }
}

MainWindow::~MainWindow()
{
    delete ui;
}

/* extract the thumbnail from raw format image
 * cache in .thumbs directory under image directory
 * return thumbnail path for display
 */
QString MainWindow::getThumbnailPath(QString imagepath) {
    /* get directory of image */
    QStringList l = imagepath.split("/");
    QString tmp_image = l.takeLast();
    QStringList c = tmp_image.split(".");
    c.pop_back();
    /* thumbnail image path */
    tmp_image = c.join(".")+".thumb.jpg";
    QString orig_path = l.join("/");
    /* create .thumbs directory if it does not exist */
    QString sel_path = orig_path+"/.thumbs/";
    if (!QDir(sel_path).exists()) {
        QDir().mkdir(sel_path);
    }
    QString sel_image = sel_path + tmp_image;
    QFileInfo t(sel_image);
    if (!t.exists()) {
        QStringList arguments;
        /* use dcraw to extract jpeg thumbnail from raw image
         * MUCH quicker than converting image to PPM.
         * Thumbnail should resemble image displayed on LCD
         * screen of camera device.
         * http://www.cybercom.net/~dcoffin/dcraw/
         * uses -e switch
         * ex: dcraw -e image.CR2
         * saves to image directory, must move to .thumbs
         */
        QString foo = "/usr/local/bin/dcraw -e "+imagepath+"";
        arguments << "-c" << foo;
        QProcess* process = new QProcess(this);
        process->start("/bin/sh", arguments);
        process->waitForFinished();
        QString tmp = process->readAll();
        /* move to .thumbs */
        QDir dp(orig_path);
        dp.rename(orig_path+"/"+tmp_image,sel_image);
    }
    return (sel_image);
}

/* load image from file into QPixmap and display on QGraphicsView */
void MainWindow::loadImage(QString imagepath,bool doStar) {

    QPixmap pm(imagepath);

    /* if select say so on status bar */
    QString sayselect = "";
    if (doStar) sayselect=" {select}";

    /* put horizontal and vertical pixels on status bar */
    QString foo = QString("%1,%2%3").arg(pm.width()).arg(pm.height()).arg(sayselect);
    statusBar()->showMessage(
        tr("burplex (%1) ")
            .arg(foo));

    gs->clear();

    /* load star from resource file
     * and add to scene
     */
    gs->addPixmap(pm);
    if (doStar) {
        QPixmap star(":/star.png");
        gs->addPixmap(star);
    }
    /* resize scene to fit image
     * this will prevent overlapping images
     * in QGraphicsView
     */
    gs->setSceneRect(pm.rect());
    gv->setScene(gs);
    gv->update();

    /* resize QGraphicsView to fit window size */
    if ((gs->width() > gv->width()) || (gs->height() > gv->height())) {
        gv->fitInView(gv->sceneRect(),Qt::KeepAspectRatio);
    } else {
        gv->resetTransform();
    }
}

/* get path to .selects file */
QString MainWindow::getSelectPath(QString imagepath) {
    QStringList l = imagepath.split("/");
    l.pop_back();
    QString sel_path = l.join("/")+"/.selects";
    return sel_path;
}

/* toggle select status of current image
 * save updated select list
 * reload image to add/remove star
 */
void MainWindow::toggleSelect(QString imagepath) {
    /* turn off */
    if (selects.contains(imagepath)) {
        selects.removeAll(imagepath);
        if (isRaw) {
            /* if it's a raw image then get the thumbnail instead */
            QString thumbpath = getThumbnailPath(imagepath);
            /* load and display image */
            loadImage(thumbpath,false);
        } else {
            /* load and display image */
            loadImage(imagepath,false);
        }
    } else {
        /* turn on */
        selects.append(imagepath);
        if (isRaw) {
            /* if it's a raw image then get the thumbnail instead */
            QString thumbpath = getThumbnailPath(imagepath);
            /* load and display image with a star */
            loadImage(thumbpath,true);
        } else {
            /* load and display image with a star */
            loadImage(imagepath,true);
        }
    }
    /* save updated select list */
    saveSelects();
}

/* save select list to image directory
 * selectpath is global var
 */
void MainWindow::saveSelects(void) {
    QFile file(selectpath);
    file.open(QIODevice::WriteOnly);
    QTextStream s(&file);
    for (int i=0;i<selects.size();++i) {
        s << selects.at(i) << "\n";
    }
    file.close();

    /* make sure we have the file
     * chk user perms
     */
    QFileInfo ds(selectpath);
    if (!ds.exists()) {
        QMessageBox msgBox;
        msgBox.setText("Cannot open selects file for writing. Check permissions! " + selectpath);
        msgBox.exec();
    }
}

/* load list of selects from .selects file in image path
 * selectpath is global var
 */
void MainWindow::loadSelects(void) {
    selects.clear();
    QFile file(selectpath);
    if (file.exists()) {
        file.open(QIODevice::ReadOnly);
        QTextStream in(&file);
        while(!in.atEnd()) {
            QString line = in.readLine();
            selects.append(line);
        }
        file.close();
    }
}

/* capture key press event
 * Insert key toggles select status
 * 'path' is global var
 */
void MainWindow::keyPressEvent( QKeyEvent *e ) {
    if (path!="") {
        QFileInfo dp(path);
        if (dp.isFile()) {
            switch( e->key() ) {

            /* key options
             * Qt::Key_Escape
             * Qt::Key_Tab
             * Qt::Key_Backtab
             * Qt::Key_Backspace
             * Qt::Key_Return
             * Qt::Key_Enter
             * Qt::Key_Insert
             * Qt::Key_Delete
             * Qt::Key_Pause
             */

                case Qt::Key_Insert:
                    toggleSelect(path);
                    break;
                default:
                    break;
            }
         }
     }
 }

//bye
